# coding=utf-8
"""
This module defines constants which are used in the API.
"""

project_version = ['0', '1', '2', 'b1']
"""
Current project version
"""

float_precision = 0.001  # 1 mm
"""
precision for float comparison
"""
